/*       +----------------------------------------------------------+
             TITLE:  RTL8019AS ethernet routines for 8051 and      
                     Keil C51 v7.00 port of Adam Dunkels' uIP v0.9 
          FILENAME:  etherdev.c                                    
          REVISION:  VER 0.0                                       
          REV.DATE:  21-01-05                                      
           ARCHIVE:                                                
            AUTHOR:  Copyright (c) 2005, Murray R. Van Luyn.       
            Refer to National Semiconductor DP8390 App Note 874.   
         +----------------------------------------------------------+        */

/* +---------------------------------------------------------------------+
      M16C-port by Jens Mundhenke, 16.05.05                             
      - The NIC is used in Memory Expansion Mode                        
      - The NIC is connected with D0-15 and A0-9. A0 is connected       
         to A1 of M16C, A1 to A2 ...                                    
      - M16C is in 16bit-mode and in memory expansion mode              
      - hardware.h defines use of ports and so on                       
          - port 6.1 is used for reset (ISA_RESET)                      
      	- INT2 is used for interrupt (ETH_INT)                           
      	- ISA_BASE is first address of used chip select of M16C          
      	  It is used with 3 wait states (for other reasons)              
                                                                        
   +---------------------------------------------------------------------+  */

/*  +---------------------------------------------------------------------+
      THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS  
      OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED  
      WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
      ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY    
      DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
      DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE  
      GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS      
      INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,       
      WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING          
      NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
      SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.       
    +---------------------------------------------------------------------+  */

#include <stdio.h>
#include "hardware.h"
#include "etherdev.h"

/*  +---------------------------------------------------------------------+
                                Private defines.                         
    +---------------------------------------------------------------------+  */
// software uses 16bit-access to DMA-register, if defined
#define ETH_16BIT

// Packet transmit & receive buffer configuration
#define ETH_TX_PAGE_START  0x40    // 0x4000 Tx buffer is  6 * 256 = 1536 bytes
#define ETH_RX_PAGE_START  0x46    // 0x4600 Rx buffer is 26 * 256 = 6656 bytes
#define ETH_RX_PAGE_STOP   0x60    // 0x6000

#define ETH_MIN_PACKET_LEN 0x3C

/*  +---------------------------------------------------------------------+
                           Private Function Prototypes                   
    +---------------------------------------------------------------------+  */
//static void etherdev_reg_write(unsigned char reg, unsigned char wr_data);


/*  +---------------------------------------------------------------------+
                              Private Macro Defines                      
    +---------------------------------------------------------------------+  */
// Manipulate PS1 & PS0 in CR to select RTL8019AS register page.
#define ETHERDEV_SELECT_REG_PAGE(page)                                      \
          do                                                                \
          {                                                                 \
              etherdev_reg_write(CR, etherdev_reg_read(CR) & ~(PS1 | PS0)); \
              etherdev_reg_write(CR, etherdev_reg_read(CR) | (page << 6));  \
          } while(0)


/*  +---------------------------------------------------------------------+
                         Private Function Implementation                 
    +---------------------------------------------------------------------+  */

/*  +---------------------------------------------------------------------+
                                                                         
                              etherdev_reg_write()                       
                                                                         
                                                                         
    +---------------------------------------------------------------------+  */
#define etherdev_reg_write(reg, data)	(*(unsigned char far *)(ETH_REG(reg))) = (unsigned char)(data)
#define etherdev_reg_write16(reg, data)	(*(unsigned int far *)(ETH_REG(reg))) = (unsigned int)(data)

/*  +---------------------------------------------------------------------+
                                                                         
                              etherdev_reg_read()                        
                                                                         
     For unknown reasons the macro implemantion is not working in every  
     case. So we have functions for the single calls, but a macro for    
     the DMA-access - the most often used transfer                       
                                                                         
    +---------------------------------------------------------------------+  */
unsigned char etherdev_reg_read(unsigned char reg)
{
  return ( (*(unsigned char far *)(ETH_REG(reg)) ) );
}

unsigned int etherdev_reg_read16f(unsigned char reg)
{
  return ( (*(unsigned int far *)(ETH_REG(reg)) ) );
}

// macro is not working in every case - but why?
#define etherdev_reg_read16(reg)	(*(volatile unsigned int far *)(ETH_REG(reg)))


/*  +---------------------------------------------------------------------+
                         Public Function Implementation                  
    +---------------------------------------------------------------------+  */

/*  +---------------------------------------------------------------------+
                                                                         
                                etherdev_init()                          
                                                                         
      Returns: 1 on success, 0 on failure.                               
      Refer to National Semiconductor DP8390 App Note 874, July 1993.    
                                                                         
    +---------------------------------------------------------------------+  */
int etherdev_init(void)
{
    // Configure RTL8019AS ethernet controller.
    unsigned int i;

    // Reset NIC on ISA-bus
    ISA_RESETD = 1;	// enable output-pin
    ISA_RESET = 1;	// reset
    for (i=0; i<100; i++);
    ISA_RESET = 0;	// ready with reset
    for (i=0; i<10000; i++);

    // enable interrupt on predefined input
    ETH_INTPORTD = 0;		// set int port to input
    ETH_IFSR = 0;		// int only on one edge
    ETH_IC = 0x10 | ETH_PRIO;	// enable int on rising edge

    // Select RTL8019AS register page 0.
    ETHERDEV_SELECT_REG_PAGE(0);

    // Check if RTL8019AS fully reset.
    if(!(etherdev_reg_read(ISR) & RST))
    {
        return 0;
    }

    // Select RTL8019AS register page 3.
    ETHERDEV_SELECT_REG_PAGE(3);

    // Temporarily disable config register write protection.
    etherdev_reg_write(_9346CR, EEM1 | EEM0);

    // Enable interrupt
    etherdev_reg_write(CONFIG1, IRQEN);

    // Disable boot ROM & select 10BaseT with TP/CX auto-detect.
    etherdev_reg_write(CONFIG2, BSELB);

    // Select half-duplex, awake, power-up & LED_TX/ LED_RX/ LED_LNK behaviour.
    etherdev_reg_write(CONFIG3, LEDS0);

    // Re-enable config register write protection.
    etherdev_reg_write(_9346CR, 0x00);

    // Select RTL8019AS register page 0.
    ETHERDEV_SELECT_REG_PAGE(0);

    // Stop RTL8019AS, select page 0 and abort DMA operation.
    etherdev_reg_write(CR, RD2 | STP);

    // Initialise data configuration register.
    // FIFO threshold 8 bytes, no loopback, don't use auto send packet.
#ifdef ETH_16BIT
    etherdev_reg_write(DCR, FT1 | LS | WTS);
#else
    etherdev_reg_write(DCR, FT1 | LS);
#endif

    // Reset remote byte count registers.
    etherdev_reg_write(RBCR0, 0x00);
    etherdev_reg_write(RBCR1, 0x00);

    // Receive configuration register to monitor mode.
    etherdev_reg_write(RCR, MON);

    // Initialise transmit configuration register to loopback internally.
    etherdev_reg_write(TCR, LB0);

    // Clear interrupt status register bits by writing 1 to each.
    etherdev_reg_write(ISR, 0xFF);

    // Enable receive interrupt in mask register.
    etherdev_reg_write(IMR, PRX);

    // Set transmit page start.
    etherdev_reg_write(TPSR, ETH_TX_PAGE_START);

    // Set receive buffer page start.
    etherdev_reg_write(PSTART, ETH_RX_PAGE_START);

    // Initialise last receive buffer read pointer.
    etherdev_reg_write(BNRY, ETH_RX_PAGE_START);

    // Set receive buffer page stop.
    etherdev_reg_write(PSTOP, ETH_RX_PAGE_STOP);

    // Select RTL8019AS register page 1.
    etherdev_reg_write(CR, RD2 | PS0 | STP);

    // Initialise current packet receive buffer page pointer
    etherdev_reg_write(CURR, ETH_RX_PAGE_START);

    // Set physical address
    etherdev_reg_write(PAR0, UIP_ETHADDR0);
    etherdev_reg_write(PAR1, UIP_ETHADDR1);
    etherdev_reg_write(PAR2, UIP_ETHADDR2);
    etherdev_reg_write(PAR3, UIP_ETHADDR3);
    etherdev_reg_write(PAR4, UIP_ETHADDR4);
    etherdev_reg_write(PAR5, UIP_ETHADDR5);

    // Select RTL8019AS register page 0 and abort DMA operation.
    etherdev_reg_write(CR, RD2 | STP);

    // Restart RTL8019AS.
    etherdev_reg_write(CR, RD2 | STA);

    // Initialise transmit configuration register for normal operation.
    etherdev_reg_write(TCR, 0x00);

    // Receive configuration register to accept broadcast packets.
    etherdev_reg_write(RCR, AB);

    return 1;
}


/*  +---------------------------------------------------------------------+
                                                                         
                               etherdev_send()                           
                                                                         
     Send the packet in the uip_buf and uip_appdata buffers using the    
     RTL8019AS ethernet card.                                            
                                                                         
    +---------------------------------------------------------------------+  */
void etherdev_send(void)
{
    unsigned int i;
#ifdef ETH_16BIT
    unsigned int far *ptr = (unsigned int far *)uip_buf;
#else
    unsigned char far *ptr = uip_buf;
#endif

    // Setup for DMA transfer from uip_buf & uip_appdata buffers to RTL8019AS.

    // Select RTL8019AS register page 0 and abort DMA operation.
    etherdev_reg_write(CR, RD2 | STA);

    // Wait until pending transmit operation completes.
    while(etherdev_reg_read(CR) & TXP) continue;

    // Clear remote DMA complete interrupt status register bit.
    etherdev_reg_write(ISR, RDC);

    // Set remote DMA start address registers to indicate where to load packet.
    etherdev_reg_write(RSAR0, 0x00);
    etherdev_reg_write(RSAR1, ETH_TX_PAGE_START);

    // Set remote DMA byte count registers to indicate length of packet load.
    etherdev_reg_write(RBCR0, (unsigned char)(uip_len & 0xFF));
    etherdev_reg_write(RBCR1, (unsigned char)(uip_len >> 8));

    // Initiate DMA transfer of uip_buf & uip_appdata buffers to RTL8019AS.
    etherdev_reg_write(CR, RD1 | STA);

    // DMA transfer packet from uip_buf & uip_appdata to RTL8019AS local
    // transmit buffer memory.
#ifdef ETH_16BIT
    for(i = 0; i < uip_len; i+=2)
    {
        if(i == 40 + UIP_LLH_LEN) {
            ptr = (unsigned int far *)uip_appdata;
        }
        etherdev_reg_write16(RDMA, *ptr++);
    }
#else
    for(i = 0; i < uip_len; i++)
    {
        if(i == 40 + UIP_LLH_LEN)
        {
            ptr = (unsigned char far *)uip_appdata;
        }

        etherdev_reg_write(RDMA, *ptr++);
    }
#endif
    // Wait until remote DMA operation complete.
    while(!(etherdev_reg_read(ISR) & RDC)) continue;

    // Abort/ complete DMA operation.
    etherdev_reg_write(CR, RD2 | STA);

    // Set transmit page start to indicate packet start.
    etherdev_reg_write(TPSR, ETH_TX_PAGE_START);

    // Ethernet packets must be > 60 bytes, otherwise are rejected as runts.
    if(uip_len < ETH_MIN_PACKET_LEN)
    {
        uip_len = ETH_MIN_PACKET_LEN;
    }

    // Set transmit byte count registers to indicate packet length.
    etherdev_reg_write(TBCR0, (unsigned char)(uip_len & 0xFF));
    etherdev_reg_write(TBCR1, (unsigned char)(uip_len >> 8));

    // Issue command for RTL8019AS to transmit packet from it's local buffer.
    etherdev_reg_write(CR, RD2 | TXP | STA);

    // MJ clear all interrupts to reset INT-line
//    etherdev_reg_write(ISR, 0xff);

    return;
}


/*  +---------------------------------------------------------------------+
                                                                         
                               etherdev_read()                           
                                                                         
     Poll the RTL8019AS ethernet device for an available packet.         
     This function will read an entire IP packet into the uip_buf.       
     If there is no packer in the NIC-buffer, it will return with        
     the return value 0. Otherwise, when a full packet has been read     
     into the uip_buf buffer, the length of the packet is returned.      
                                                                         
    +---------------------------------------------------------------------+  */

unsigned int etherdev_read(void)
{
    unsigned int len = 0;

    // Check if there is a packet in the rx buffer.
    if(etherdev_reg_read(ISR) & PRX)
    {
        // Check if the rx buffer has overflowed.
        if(etherdev_reg_read(ISR) & OVW)
        {
            char retransmit = 0;

            // If the receive buffer ring has overflowed we dump the whole
            // thing and start over. There is no way of knowing whether the
            // data it contains is uncorrupted, or will cause us grief.

            // Stop RTL8019AS and abort DMA operation.
            etherdev_reg_write(CR, RD2 | STP);

            // Reset remote byte count registers.
            etherdev_reg_write(RBCR0, 0x00);
            etherdev_reg_write(RBCR1, 0x00);

            // Wait for controller to halt after any current tx completes.
            while(!(etherdev_reg_read(ISR) & RST)) continue;

            // Check whether currently transmitting a packet.
            if(etherdev_reg_read(CR) & TXP)
            {
                // If neither a successful transmission nor a tx abort error
                // has occured, then flag current tx packet for resend.
                if(!((etherdev_reg_read(ISR) & PTX)
                                          || (etherdev_reg_read(ISR) & TXE)))
                {
                    retransmit = 1;
                }
            }

            // Set transmit configuration register to loopback internally.
            etherdev_reg_write(TCR, LB0);

            // Restart the RTL8019AS.
            etherdev_reg_write(CR, RD2 | STA);

            // Re-initialise last receive buffer read pointer.
            etherdev_reg_write(BNRY, ETH_RX_PAGE_START);

            // Select RTL8019AS register page 1.
            ETHERDEV_SELECT_REG_PAGE(1);

            // Re-initialise current packet receive buffer page pointer.
            etherdev_reg_write(CURR, ETH_RX_PAGE_START);

            // Select RTL8019AS register page 0.
            ETHERDEV_SELECT_REG_PAGE(0);

            // Clear rx buffer overflow & packet received interrupt flags.
            etherdev_reg_write(ISR, PRX | OVW);

            // Re-itialise transmit configuration reg for normal operation.
            etherdev_reg_write(TCR, 0x00);

            if(retransmit)
            {
                // Retransmit packet in RTL8019AS local tx buffer.
                etherdev_reg_write(CR, RD2 | TXP | STA);
            }
        }
        else // Rx buffer has not overflowed, so read a packet into uip_buf.
        {
            unsigned int i;
            unsigned char next_rx_packet;
            unsigned char current;

            // Retrieve packet header. (status, next_ptr, length_l, length_h)

            // Clear remote DMA complete interrupt status register bit.
            etherdev_reg_write(ISR, RDC);

            // Set remote DMA start address registers to packet header.
            etherdev_reg_write(RSAR0, 0x00);
            etherdev_reg_write(RSAR1, etherdev_reg_read(BNRY));

            // Set remote DMA byte count registers to packet header length.
            etherdev_reg_write(RBCR0, 0x04);
            etherdev_reg_write(RBCR1, 0x00);

            // Initiate DMA transfer of packet header.
            etherdev_reg_write(CR, RD0 | STA);

#ifdef ETH_16BIT
            // Save next packet pointer.
            // Works only with the read function, not with macro - but why???
            next_rx_packet = etherdev_reg_read16f(RDMA) >> 8;

            // Retrieve packet data length
            len =  etherdev_reg_read16(RDMA);
#else
            // Drop packet status. We don't use it.
            etherdev_reg_read(RDMA);

            // Save next packet pointer.
            next_rx_packet = etherdev_reg_read(RDMA);

            // Retrieve packet data length
            len =  etherdev_reg_read(RDMA);
            i   =  etherdev_reg_read(RDMA);
            i   <<= 8;
            len += i;
#endif
            // Subtract CRC bytes.
            len -= 4;

            // Wait until remote DMA operation completes.
            while(!(etherdev_reg_read(ISR) & RDC)) continue;

            // Abort/ complete DMA operation.
            etherdev_reg_write(CR, RD2 | STA);

            // Retrieve packet data.

            // Check if incoming packet will fit into rx buffer.
            if(len <= UIP_BUFSIZE)
            {
                // Clear remote DMA complete interrupt status register bit.
                etherdev_reg_write(ISR, RDC);

                // Set remote DMA start address registers to packet data.
                etherdev_reg_write(RSAR0, 0x04);
                etherdev_reg_write(RSAR1, etherdev_reg_read(BNRY));

                // Set remote DMA byte count registers to packet data length.
                etherdev_reg_write(RBCR0, (unsigned char)(len & 0xFF));
                etherdev_reg_write(RBCR1, (unsigned char)(len >> 8));

                // Initiate DMA transfer of packet data.
                etherdev_reg_write(CR, RD0 | STA);

                // Read packet data directly into uip_buf.
#ifdef ETH_16BIT
                for(i = 0; i < len; i+=2)
                    *(unsigned int far *)(uip_buf + i) = etherdev_reg_read16(RDMA);
#else
                for(i = 0; i < len; i++)
                    *(uip_buf + i) = etherdev_reg_read(RDMA);
#endif
                // Wait until remote DMA operation complete.
                while(!(etherdev_reg_read(ISR) & RDC)) continue;

                // Abort/ complete DMA operation.
                etherdev_reg_write(CR, RD2 | STA);

            }
            else
            {
                // Incoming packet too big, so dump it.
                len = 0;
            }

            // Advance boundary pointer to next packet start.
            etherdev_reg_write(BNRY, next_rx_packet);

            // Select RTL8019AS register page 1.
            ETHERDEV_SELECT_REG_PAGE(1);

            // Retrieve current receive buffer page
            current = etherdev_reg_read(CURR);

            // Select RTL8019AS register page 0.
            ETHERDEV_SELECT_REG_PAGE(0);

            // Check if last packet has been removed from rx buffer.
            if(next_rx_packet == current)
            {
                // Clear packet received interrupt flag.
                etherdev_reg_write(ISR, PRX);
            }
        }
    }

    return len;
}



/*  +---------------------------------------------------------------------+
                                                                         
                               ETH_INT()                                 
                                                                         
     Interrupt function called from NIC-interruptline, only PRX-int is   
     active. Interrupt-support is still very simple                      
                                                                         
    +---------------------------------------------------------------------+  */
void ETH_INT(void)
{
  char isr;

  isr = etherdev_reg_read(ISR);

  if ( isr & PRX )
    uipmain_read();
}

